from abc import ABC, abstractmethod


class AbstractFormatter(ABC):
    """An abstract formatter."""

    def __init__(self) -> None:
        super().__init__()

    @abstractmethod
    def format(self, data):
        """
        :param data: The data to format.
        :return: The new format of the data.
        """



from bobocep.decider.dist_decider import DistDecider
from bobocep.setup.distributed.incoming.bobo_dist_incoming import \
    BoboDistIncoming
from bobocep.setup.distributed.outgoing.bobo_dist_outgoing import \
    BoboDistOutgoing
from bobocep.setup.task.bobo_task_thread import \
    BoboTaskThread


class BoboDistManager:
    """A manager for incoming and outgoing data that helps multiple
    :code:`bobocep` instances synchronise data between each other using an
    external message queue system.

    :param decider: The decider to which synchronisation will occur.
    :type decider: DistDecider

    :param exchange_name: The exchange name to connect to on the external
                          message queue system.
    :type exchange_name: str

    :param user_id: The user ID to use on the external message queue system.
    :type user_id: str

    :param host_name: The host name of the external message queue system.
    :type host_name: str

    :param delay: The delay for internal BoboTask threads, in seconds.
    :type delay: float
    """

    def __init__(self,
                 decider: DistDecider,
                 exchange_name: str,
                 user_id: str,
                 host_name: str,
                 delay: float) -> None:
        super().__init__()

        self.incoming = BoboDistIncoming(decider=decider,
                                         exchange_name=exchange_name,
                                         user_id=user_id,
                                         host_name=host_name)

        self.outgoing = BoboDistOutgoing(decider=decider,
                                         exchange_name=exchange_name,
                                         user_id=user_id,
                                         host_name=host_name)

        self.incoming.subscribe(self.outgoing)
        self.outgoing.subscribe(self.incoming)

        self._incoming_thread = BoboTaskThread(self.incoming, delay)
        self._outgoing_thread = BoboTaskThread(self.outgoing, delay)

    def start(self):
        """Run the manager threads."""

        self._incoming_thread.start()
        self._outgoing_thread.start()

    def cancel(self):
        """Cancel the manager threads."""

        self._incoming_thread.cancel()
        self._outgoing_thread.cancel()



from typing import List

from bobocep.rules.events.bobo_event import BoboEvent
from bobocep.rules.events.composite_event import CompositeEvent
from bobocep.rules.events.histories.bobo_history import BoboHistory
from bobocep.rules.predicates.windows.bobo_predicate_window import \
    BoboPredicateWindow


class WindowFixed(BoboPredicateWindow):
    """A predicate that evaluates using a fixed time window.

    :param lower_ns: The lower bound of the fixed window, as Epoch timestamp
                     with nanosecond precision.
    :type lower_ns: int

    :param upper_ns: The upper bound of the fixed window, as Epoch timestamp
                     with nanosecond precision.
    :type upper_ns: int

    :raises RuntimeError: The lower bound is greater than or equal to the upper
                          bound.
    """

    def __init__(self, lower_ns: int, upper_ns: int) -> None:
        super().__init__()

        if lower_ns >= upper_ns:
            raise RuntimeError

        self._lower_ns = lower_ns
        self._upper_ns = upper_ns

    def evaluate(self,
                 event: BoboEvent,
                 history: BoboHistory,
                 recents: List[CompositeEvent]) -> bool:
        return self._lower_ns <= event.timestamp <= self._upper_ns



from bobocep.decider.buffers.abstract_buffer import AbstractBuffer
from bobocep.decider.buffers.match_event import MatchEvent
from bobocep.decider.versions.run_version import RunVersion
from bobocep.rules.events.bobo_event import BoboEvent
from bobocep.rules.events.histories.bobo_history import BoboHistory


class SharedVersionedMatchBuffer(AbstractBuffer):
    """
    An event buffer based on the Shared Versioned Match Buffer proposed by
    Agrawal et al. (2008) in their paper "Efficient Pattern Matching over
    Event Streams".

    The buffer stores events for partially completed runs.
    A BoboEvent instance is stored within a MatchEvent instance that provides
    a means of linking BoboEvents that can be used by one or more runs.
    The next event in a MatchEvent link points to an earlier event accepted by
    a run.
    Traversing the next links will identify all events for a given run.
    """

    NFA_NAME = "nfa_name"
    LABEL = "label"
    MATCH_EVENT = "match_event"
    RUN_ID = "run_id"
    VERSION = "version"
    EVENT_ID = "event_id"
    EVENTS = "events"
    LAST = "last"

    def __init__(self) -> None:
        super().__init__()

        # BoboNFA Name -> BoboState Label -> BoboEvent ID -> MatchEvent
        self._eve = {}

        # BoboNFA Name -> BoboRun ID -> RunVersion -> Last MatchEvent
        self._ver = {}

    @staticmethod
    def _get_or_create_subdict(d: dict, key: str) -> dict:
        """
        Gets a nested dict inside an existing dict, or creates one there if one
        does not exist.

        :param d: The dict in which to create a dict.
        :type d: dict

        :param key: The key that will point to the new dict.
        :type key: str

        :return: The nested dict.
        """
        if key not in d:
            d[key] = {}

        return d.get(key)

    def get_event(self,
                  nfa_name: str,
                  state_label: str,
                  event_id: str,
                  default=None) -> BoboEvent:
        """
        Get a BoboEvent instance from the buffer.

        :param nfa_name: The NFA name,
        :type nfa_name: str

        :param state_label: The state label.
        :type state_label: str

        :param event_id: The event ID.
        :type event_id: str

        :param default: The default value, defaults to None.
        :type default: any, optional

        :return: The BoboEvent instance,
                 or default value if an event is not found.
        """

        try:
            return self._eve[nfa_name][state_label][event_id].event

        except KeyError:
            return default

    def put_event(self,
                  nfa_name: str,
                  run_id: str,
                  version: str,
                  state_label: str,
                  event: BoboEvent,
                  new_run_id: str = None,
                  new_version: str = None) -> MatchEvent:
        """
        Puts a BoboEvent instance into the buffer.

        :param nfa_name: The BoboNFA instance name.
        :type nfa_name: str

        :param run_id: The run ID with which to associate the event.
        :type run_id: str

        :param version: The run version of the run with which to associate the
                        event.
        :type version: str

        :param state_label: The label of the state .
        :type state_label: str

        :param event: The event to add to the buffer.
        :type event: BoboEvent

        :param new_run_id: The new run ID with which to associate the event,
                           so that the other run ID will point to this one,
                           defaults to None.
        :type new_run_id: str, optional

        :param new_version: The new run version with which to associate the
                            event, so that the other version will point to this
                            one, defaults to None.
        :type new_version: str, optional

        :return: A MatchEvent instance containing the BoboEvent instance.
        """

        # get match events for this nfa, keyed by the bobo event they represent
        nfa_labels = self._get_or_create_subdict(self._eve, nfa_name)
        nfa_events = self._get_or_create_subdict(nfa_labels, state_label)

        # look for bobo event in buffer, or add a new match event for it
        # if a match event isn't found
        if event.id not in nfa_events:
            nfa_events[event.id] = MatchEvent(
                nfa_name=nfa_name,
                label=state_label,
                event=event)

        new_match_event = nfa_events[event.id]

        # get last event under the original run ID and version
        # before (maybe) updating run ID and version
        last_match_event = self.get_last_event(
            nfa_name=nfa_name,
            run_id=run_id,
            version=version)

        if new_run_id is not None:
            run_id = new_run_id

        if new_version is not None:
            version = new_version

        # point the (maybe new) run ID and version to the next match event
        nfa_runs = self._get_or_create_subdict(self._ver, nfa_name)
        run_versions = self._get_or_create_subdict(nfa_runs, run_id)
        run_versions[version] = new_match_event

        # point new match event to the last match event, if it exists
        if last_match_event is not None:
            # points to the last match event
            new_match_event.add_pointer_next(
                version=version,
                label=last_match_event.label,
                event_id=last_match_event.event.id)

            # points backwards to the new match event
            last_match_event.add_pointer_previous(
                version=version,
                label=new_match_event.label,
                event_id=new_match_event.event.id)
        else:
            # adds pointer to nothing, so run is still linked to event
            new_match_event.add_pointer_next(version=version)

        return new_match_event

    def remove_version(self, nfa_name: str, version: str) -> None:
        """
        Removes a run version from all of the match events in the buffer.

        :param nfa_name: The name of the BoboNFA instance.
        :type nfa_name: str

        :param version: The run version.
        :type version: str
        """

        # remove version from runs
        nfa_runs = self._ver.get(nfa_name)

        if nfa_runs is not None:
            for nfa_run in tuple(nfa_runs.values()):
                for run_version in tuple(nfa_run.keys()):
                    if version == run_version:
                        nfa_runs.pop(version, None)

        # remove pointers in events, and maybe event itself
        nfa_labels = self._eve.get(nfa_name)

        if nfa_labels is not None:
            for match_events in tuple(nfa_labels.values()):
                for match_event in tuple(match_events.values()):
                    match_event.remove_all_pointers(version)

                    # drop more pointers, drop event
                    if not match_event.has_pointers():
                        self._eve \
                            .get(nfa_name) \
                            .get(match_event.label) \
                            .pop(match_event.event.id, None)

    def get_last_event(self,
                       nfa_name: str,
                       run_id: str,
                       version: str,
                       default=None) -> MatchEvent:
        """
        Gets the last match event.

        :param nfa_name: The BoboNFA instance name.
        :type nfa_name: str

        :param run_id: The run ID.
        :type run_id: str

        :param version: The run version.
        :type version: str

        :param default: A value to return if no match event is found,
                        defaults to None.
        :type default: any, optional

        :return: The last match event,
                 or a default value if one does not exist.
        """

        try:
            return self._ver[nfa_name][run_id][version]

        except KeyError:
            return default

    def get_all_events(self,
                       nfa_name: str,
                       run_id: str,
                       version: RunVersion) -> BoboHistory:
        """
        Gets all events associated with a run and compiles them into a
        BoboHistory instance.

        :param nfa_name: The BoboNFA instance name.
        :type nfa_name: str

        :param run_id: The run ID.
        :type run_id: str

        :param version: The run version.
        :type version: RunVersion

        :return: A BoboHistory instance with all of the events in it.
        """

        all_events = {}
        current_level = 0
        current_incr = 0
        current_version = version.get_version_as_str()

        # start with the latest match event
        current_event = self.get_last_event(
            nfa_name=nfa_name,
            run_id=run_id,
            version=current_version)

        while True:
            if current_event is not None:
                # add event to dict, keyed under the label name
                if current_event.label not in all_events:
                    all_events[current_event.label] = []
                all_events[current_event.label].insert(0, current_event.event)

                # get next match event using current version
                next_event = self._get_next_event(
                    event=current_event,
                    nfa_name=nfa_name,
                    version_str=current_version)

                # no event found under current version
                if next_event is None:
                    # get previous version by decreasing increment
                    current_incr += 1
                    current_version = \
                        version.get_previous_version_as_str(
                            decrease_level=current_level,
                            decrease_incr=current_incr)

                    # get previous version by decreasing level
                    if current_version is None:
                        current_level += 1
                        current_incr = 0
                        current_version = \
                            version.get_previous_version_as_str(
                                decrease_level=current_level,
                                decrease_incr=current_incr)

                        # no previous version, stop search
                        if current_version is None:
                            break

                    # attempt to find next event with new version
                    next_event = self._get_next_event(
                        event=current_event,
                        nfa_name=nfa_name,
                        version_str=current_version)

                    if next_event is None:
                        break

                current_event = next_event
            else:
                break

        return BoboHistory(events=all_events)

    def _get_next_event(self,
                        event: MatchEvent,
                        nfa_name: str,
                        version_str: str):
        try:
            next_tuple = event.next_ids.get(version_str)
            if next_tuple is not None:
                next_event = self._eve[nfa_name][next_tuple[0]][next_tuple[1]]
            else:
                next_event = None
        except KeyError:
            next_event = None

        return next_event

    def to_dict(self) -> dict:
        """
        :return: A dict representation of the object.
        """

        eve = []
        ver = []

        for nfa_name, nfa_dict in self._eve.items():
            for event_dict in nfa_dict.values():
                for match_event in event_dict.values():
                    eve.append({
                        self.NFA_NAME: nfa_name,
                        self.MATCH_EVENT: match_event.to_dict()
                    })

        for nfa_name, nfa_dict in self._ver.items():
            for run_id, run_dict in nfa_dict.items():
                for version, match_event in run_dict.items():
                    ver.append({
                        self.NFA_NAME: nfa_name,
                        self.LABEL: match_event.label,
                        self.RUN_ID: run_id,
                        self.VERSION: version,
                        self.EVENT_ID: match_event.event.id
                    })

        return {
            self.EVENTS: eve,
            self.LAST: ver
        }



from abc import ABC, abstractmethod
from queue import Queue

from bobocep.forwarder.abstract_forwarder import AbstractForwarder
from bobocep.forwarder.forwarder_subscriber import IForwarderSubscriber
from bobocep.producer.producer_subscriber import IProducerSubscriber
from bobocep.rules.events.composite_event import CompositeEvent
from bobocep.setup.task.bobo_task import BoboTask


class BoboForwarder(AbstractForwarder,
                    BoboTask,
                    IProducerSubscriber,
                    ABC):
    """A :code:`bobocep` event forwarder that forwards CompositeEvent
    instances.

    :param max_queue_size: The maximum data queue size,
                           defaults to 0 (infinite).
    :type max_queue_size: int, optional
    """

    def __init__(self, max_queue_size: int = 0) -> None:
        super().__init__()

        self._event_queue = Queue(maxsize=max_queue_size)
        self._subs = []

    @abstractmethod
    def _handle_composite_event(self, event: CompositeEvent) -> bool:
        """
        Enables the forwarder to perform some action with the CompositeEvent
        instance before it is sent to the subscribers of the forwarder.
        If the method returns True, the CompositeEvent instance will be
        forwarded to subscribers.

        :param event: The event.
        :type event: CompositeEvent

        :return: True if event was successfully handled, False otherwise.
        """

    def _loop(self) -> None:
        while not self._event_queue.empty():
            event = self._event_queue.get_nowait()

            if event is not None:
                if self._handle_composite_event(event):
                    self._notify_success(event)
                else:
                    self._notify_failure(event)

    def on_accepted_producer_event(self, event: CompositeEvent):
        if not self._cancelled:
            self._event_queue.put_nowait(event)

    def subscribe(self, subscriber: IForwarderSubscriber) -> None:
        """
        :param subscriber: Subscribes to events from Forwarder.
        :type subscriber: IForwarderSubscriber
        """

        with self._lock:
            if not self._cancelled:
                if subscriber not in self._subs:
                    self._subs.append(subscriber)

    def unsubscribe(self, unsubscriber: IForwarderSubscriber) -> None:
        """
        :param unsubscriber: Unsubscribes to events from Forwarder.
        :type unsubscriber: IForwarderSubscriber
        """

        with self._lock:
            if unsubscriber in self._subs:
                self._subs.remove(unsubscriber)

    def _notify_success(self, event: CompositeEvent):
        for subscriber in self._subs:
            subscriber.on_forwarder_success_event(event)

    def _notify_failure(self, event: CompositeEvent):
        for subscriber in self._subs:
            subscriber.on_forwarder_failure_event(event)

    def on_rejected_producer_event(self, event: CompositeEvent) -> None:
        """"""

    def _setup(self) -> None:
        """"""

    def _cancel(self) -> None:
        """"""



from abc import ABC, abstractmethod

from bobocep.rules.events.composite_event import CompositeEvent


class IForwarderSubscriber(ABC):
    """An interface to subscribe to Forwarder events."""

    @abstractmethod
    def on_forwarder_success_event(self, event: CompositeEvent) -> None:
        """
        Events that have been successfully forwarded.

        :param event: A successful event.
        :type event: CompositeEvent
        """

    @abstractmethod
    def on_forwarder_failure_event(self, event: CompositeEvent) -> None:
        """
        Events that failed to be forwarded.

        :param event: An unsuccessful event.
        :type event: CompositeEvent
        """



from threading import RLock
from typing import Dict

from bobocep.rules.actions.bobo_action import BoboAction
from bobocep.rules.events.composite_event import CompositeEvent


class RateLimitAction(BoboAction):
    """
    Ensures that CompositeEvent instances of a given name are only produced
    at a given rate i.e. one event with name *n* in *t* seconds.
    If multiple instances are created within the rate, the subsequence
    ones are dropped i.e. the action returns :code:`False`.

    :param limit_dict: Key-value pairs, where the key is the CompositeEvent
                       name and the value is the desired rate, in seconds.
                       Defaults to an empty dict.
    :type limit_dict: Dict[str, int], optional

    :param rate_other: The rate for all other events that have not been named,
                       defaults to 0.
    :type rate_other: int, optional
    """

    def __init__(self,
                 limit_dict: Dict[str, int] = None,
                 rate_other: int = 0) -> None:
        super().__init__()

        self._limit_dict = limit_dict if limit_dict is not None else {}
        self._names = list(self._limit_dict.keys())
        self._rate_other = rate_other
        self._last = {}
        self._lock = RLock()

    def get_limits(self) -> Dict[str, int]:
        """
        :return: A dict, where the keys are the names to limit and the
                 values are the desired rates for each name, in seconds.
        """
        with self._lock:
            return self._limit_dict.copy()

    def set_limit(self, name: str, rate: int) -> None:
        """
        :param name: The CompositeEvent name.
        :type name: str

        :param rate: The rate for the name.
        :type rate: int
        """

        with self._lock:
            self._limit_dict[name] = rate

            if name not in self._names:
                self._names.append(name)

    def perform_action(self, event: CompositeEvent) -> bool:
        with self._lock:
            name = event.name
            rate = self._limit_dict[name] if name in self._names \
                else self._rate_other

            if rate > 0:
                if name not in self._last:
                    self._last[name] = 0

                accept = (event.timestamp - self._last[name]) / 1e9 >= rate

                if accept:
                    self._last[name] = event.timestamp

                return accept

            return True



from threading import Thread, Event

from bobocep.setup.task.bobo_task import BoboTask


class BoboTaskThread(Thread):
    """
    A thread to run BoboTask instances. It runs the setup task once, and
    then runs the loop task continually until the thread is cancelled.
    There is a delay between each loop.

    :param task: The task to perform.
    :type task: BoboTask

    :param delay: The delay between loops.
    :type delay: float
    """

    def __init__(self,
                 task: BoboTask,
                 delay: float):
        Thread.__init__(self)

        self._task = task
        self._delay = delay
        self._cancel = Event()

    def run(self):
        """Runs the task.

        :raises RuntimeError: Exception raised during task execution.
        """

        try:
            if not self._cancel.is_set():
                self._task.setup()

                # THREAD has been cancelled if False
                while not self._cancel.wait(self._delay):
                    self._task.loop()

                self._task.cancel()
        except Exception as e:
            raise RuntimeError("Exception raised during task execution: {}."
                               .format(e))

    def cancel(self):
        """Cancels the task."""

        self._cancel.set()



from abc import abstractmethod


class AbstractValidator:
    """An abstract validator."""

    def __init__(self) -> None:
        super().__init__()

    @abstractmethod
    def validate(self, data) -> bool:
        """
        :param data: The data to validate.
        :return: True if the data is valid, False otherwise.
        """



from queue import Queue

from bobocep.receiver.abstract_receiver import AbstractReceiver
from bobocep.receiver.formatters.primitive_event_formatter import \
    PrimitiveEventFormatter
from bobocep.receiver.receiver_subscriber import IReceiverSubscriber
from bobocep.receiver.validators.abstract_validator import \
    AbstractValidator
from bobocep.rules.events.primitive_event import PrimitiveEvent
from bobocep.setup.task.bobo_task import BoboTask


class BoboReceiver(AbstractReceiver,
                   BoboTask):
    """A :code:`bobocep` data receiver.

    :param validator: The data validator.
    :type validator: AbstractValidator

    :param formatter: The event formatter.
    :type formatter: PrimitiveEventFormatter

    :param max_queue_size: The maximum data queue size,
                           defaults to 0 (infinite).
    :type max_queue_size: int, optional
    """

    def __init__(self,
                 validator: AbstractValidator,
                 formatter: PrimitiveEventFormatter,
                 max_queue_size: int = 0) -> None:
        super().__init__()

        self._data_queue = Queue(maxsize=max_queue_size)
        self._validator = validator
        self._formatter = formatter
        self._subs = []

    def _loop(self) -> None:
        while not self._data_queue.empty():
            data = self._data_queue.get_nowait()

            if self._validator.validate(data):
                self._notify_primitive_event(
                    self._formatter.format(data))
            else:
                self._notify_invalid_data(data)

    def add_data(self, data) -> None:
        """
        Add data to the receiver.

        :param data: Data to add.
        :type data: any
        """

        if not self._cancelled:
            self._data_queue.put_nowait(data)

    def subscribe(self, subscriber: IReceiverSubscriber) -> None:
        """
        :param subscriber: Subscribes to events from Receiver.
        :type subscriber: IReceiverSubscriber
        """

        with self._lock:
            if not self._cancelled and subscriber not in self._subs:
                self._subs.append(subscriber)

    def unsubscribe(self, unsubscriber: IReceiverSubscriber) -> None:
        """
        :param unsubscriber: Unsubscribes to events from Receiver.
        :type unsubscriber: IReceiverSubscriber
        """

        with self._lock:
            if unsubscriber in self._subs:
                self._subs.remove(unsubscriber)

    def _notify_primitive_event(self, event: PrimitiveEvent) -> None:
        for subscriber in self._subs:
            subscriber.on_receiver_event(event)

    def _notify_invalid_data(self, data) -> None:
        for subscriber in self._subs:
            subscriber.on_invalid_data(data)

    def _setup(self) -> None:
        """"""

    def _cancel(self) -> None:
        """"""



from copy import copy
from typing import Dict

from bobocep.receiver.bobo_receiver import BoboReceiver
from bobocep.setup.task.bobo_task import BoboTask


class BoboNullDataGenerator(BoboTask):
    """
    A generator that periodically sends copies of some arbitrary static data
    (i.e. *null data*) to a BoboReceiver instance.

    :param receiver: The data receiver to which null events are sent.
    :type receiver: BoboReceiver

    :param null_data: The null data to send, defaults to an empty dict.
    :type null_data: Dict[str, str], optional
    """

    def __init__(self,
                 receiver: BoboReceiver,
                 null_data: Dict[str, str] = None) -> None:
        super().__init__()

        self.receiver = receiver
        self.null_data = null_data if null_data is not None else {}

    def _loop(self) -> None:
        self.receiver.add_data(copy(self.null_data))

    def _setup(self) -> None:
        """"""

    def _cancel(self) -> None:
        """"""



from queue import Queue
from typing import List

from bobocep.decider.abstract_decider import AbstractDecider
from bobocep.decider.decider_subscriber import IDeciderSubscriber
from bobocep.decider.handlers.bobo_nfa_handler import BoboNFAHandler
from bobocep.decider.handlers.nfa_handler_subscriber import \
    INFAHandlerSubscriber
from bobocep.producer.producer_subscriber import IProducerSubscriber
from bobocep.receiver.receiver_subscriber import IReceiverSubscriber
from bobocep.rules.events.bobo_event import BoboEvent
from bobocep.rules.events.composite_event import CompositeEvent
from bobocep.setup.task.bobo_task import BoboTask


class BoboDecider(AbstractDecider,
                  BoboTask,
                  IReceiverSubscriber,
                  INFAHandlerSubscriber,
                  IProducerSubscriber):
    """A :code:`bobocep` data decider.

    :param max_queue_size: The maximum data queue size,
                           defaults to 0 (infinite).
    :type max_queue_size: int, optional
    """

    HANDLERS = "handlers"

    def __init__(self, max_queue_size: int = 0) -> None:
        super().__init__()

        self._event_queue = Queue(maxsize=max_queue_size)
        self._nfa_handlers = {}
        self._subs = {}

    def _loop(self) -> None:
        if not self._event_queue.empty():
            event = self._event_queue.get_nowait()

            if event is not None:
                for nfa_handler in self._nfa_handlers.values():
                    nfa_handler.process(event)

    def add_nfa_handler(self, nfa_handler: BoboNFAHandler) -> None:
        """
        Adds a new NFA handler to the decider.

        :param nfa_handler: The NFA handler to add.
        :type nfa_handler: BoboNFAHandler

        :raises RuntimeError: NFA handler name already in use.
        """

        with self._lock:
            if not self._cancelled:
                if nfa_handler.nfa.name not in self._nfa_handlers:
                    self._nfa_handlers[nfa_handler.nfa.name] = nfa_handler
                    nfa_handler.subscribe(self)
                else:
                    raise RuntimeError("NFA handler name {} already in use."
                                       .format(nfa_handler.nfa.name))

    def get_handlers(self) -> List[BoboNFAHandler]:
        return list(self._nfa_handlers.values())

    def on_receiver_event(self, event: BoboEvent) -> None:
        if not self._cancelled:
            self._event_queue.put(event)

    def on_accepted_producer_event(self, event: CompositeEvent) -> None:
        if not self._cancelled:
            self._event_queue.put(event)

    def on_handler_final(self,
                         nfa_name: str,
                         run_id: str,
                         event: CompositeEvent) -> None:
        with self._lock:
            if not self._cancelled:
                self._notify_new_complex_event(
                    nfa_name=nfa_name,
                    event=event)

    def _notify_new_complex_event(self,
                                  nfa_name: str,
                                  event: CompositeEvent):
        if nfa_name in self._subs:
            for sub in self._subs[nfa_name]:
                sub.on_decider_complex_event(nfa_name=nfa_name, event=event)

    def subscribe(self,
                  nfa_name: str,
                  subscriber: IDeciderSubscriber) -> None:
        """
        :param nfa_name: The NFA to subscribe to.
        :type nfa_name: str

        :param subscriber: Subscribes to NFA in Decider.
        :type subscriber: IDeciderSubscriber

        :raises RuntimeError: NFA name not found.
        """

        with self._lock:
            if not self._cancelled:
                if nfa_name not in self._nfa_handlers:
                    raise RuntimeError("NFA name {} not found in handlers."
                                       .format(nfa_name))

                if nfa_name not in self._subs:
                    self._subs[nfa_name] = []

                if subscriber not in self._subs[nfa_name]:
                    self._subs[nfa_name].append(subscriber)

    def unsubscribe(self,
                    nfa_name: str,
                    unsubscriber: IDeciderSubscriber) -> None:
        """
        :param nfa_name: The NFA to unsubscribe from.
        :type nfa_name: str

        :param unsubscriber: Subscribes to NFA in Decider.
        :type unsubscriber: IDeciderSubscriber
        """

        with self._lock:
            if nfa_name in self._subs:
                if unsubscriber in self._subs[nfa_name]:
                    self._subs[nfa_name].remove(unsubscriber)

    def to_dict(self) -> dict:
        """
        :return: A dict representation of the object.
        """

        with self._lock:
            return {
                self.HANDLERS: [handler.to_dict() for handler in
                                self._nfa_handlers.values()]
            }

    def on_invalid_data(self, data) -> None:
        """"""

    def on_rejected_producer_event(self, event: CompositeEvent) -> None:
        """"""

    def on_handler_transition(self,
                              nfa_name: str,
                              run_id: str,
                              state_name_from: str,
                              state_name_to: str,
                              event: BoboEvent) -> None:
        """"""

    def on_handler_clone(self,
                         nfa_name: str,
                         run_id: str,
                         state_name: str,
                         event: BoboEvent) -> None:
        """"""

    def on_handler_halt(self,
                        nfa_name: str,
                        run_id: str) -> None:
        """"""

    def _setup(self) -> None:
        """"""

    def _cancel(self) -> None:
        """"""



from abc import abstractmethod
from typing import List

from bobocep.rules.events.bobo_event import BoboEvent
from bobocep.rules.events.composite_event import CompositeEvent
from bobocep.rules.events.histories.bobo_history import BoboHistory
from bobocep.rules.predicates.abstract_predicate import AbstractPredicate


class BoboPredicate(AbstractPredicate):
    """A :code:`bobocep` predicate."""

    def __init__(self) -> None:
        super().__init__()

    @abstractmethod
    def evaluate(self,
                 event: BoboEvent,
                 history: BoboHistory,
                 recents: List[CompositeEvent]) -> bool:
        """Evaluates the predicate.

        :param event: An event.
        :type event: BoboEvent

        :param history: A history of events.
        :type history: BoboHistory

        :param recents: Recently accepted complex events of the corresponding
                        automaton.
        :type recents: List[CompositeEvent]

        :return: True if the predicate evaluates to True,
                 False otherwise.
        :rtype: bool
        """



from inspect import isfunction, signature, ismethod
from typing import Callable, List

from bobocep.rules.events.bobo_event import BoboEvent
from bobocep.rules.events.composite_event import CompositeEvent
from bobocep.rules.events.histories.bobo_history import BoboHistory
from bobocep.rules.predicates.bobo_predicate import BoboPredicate


class BoboPredicateCallable(BoboPredicate):
    """A predicate that evaluates using a custom function or method.

    :param call: A callable function that is used to evaluate the
                 predicate. It must have exactly 3 parameters in its
                 signature and return a bool response.
    :type call: Callable

    :raises RuntimeError: Callable is not a function or method.
    :raises RuntimeError: Callable does not have exactly 3 parameters
                          in its signature.
    """

    PARAMETERS = 3

    def __init__(self, call: Callable) -> None:
        super().__init__()

        if not (isfunction(call) or ismethod(call)):
            raise RuntimeError("Callable must be a function or method, "
                               "found {}.".format(type(call)))

        if (len(signature(call).parameters)) != self.PARAMETERS:
            raise RuntimeError("Callable must have exactly {} parameters "
                               "in its signature.".format(self.PARAMETERS))

        self._call = call

    def evaluate(self,
                 event: BoboEvent,
                 history: BoboHistory,
                 recents: List[CompositeEvent]) -> bool:
        return self._call(event, history, recents)



from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    long_description = readme_file.read()

requirements = []
setup_requirements = []
test_requirements = []

setup(
    author="r3w0p",
    author_email='',
    name='bobocep',
    version='0.26.0',
    description="A fault-tolerant complex event processing engine designed "
                "for edge computing in IoT systems.",
    long_description=long_description,
    long_description_content_type='text/x-rst',
    license="MIT",
    keywords='bobocep',
    url='https://github.com/r3w0p/bobocep',

    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.7',
    ],

    install_requires=requirements,
    setup_requires=setup_requirements,
    tests_require=test_requirements,
    packages=find_packages(include=['bobocep', 'bobocep.*']),
    test_suite='tests',
    zip_safe=False,
)



from typing import List

from bobocep.rules.actions.bobo_action import BoboAction
from bobocep.rules.actions.multi.multi_action import MultiAction
from bobocep.rules.events.composite_event import CompositeEvent


class SequentialAction(MultiAction):
    """An action that enables the sequential execution of multiple actions
    within a single BoboAction instance.

    :param actions: The actions to execute, in list order.
    :type actions: List[BoboAction]

    :param all_success: All actions must be successful for perform_action to
                        return True, defaults to True.
    :type all_success: bool, optional

    :param early_stop: Will immediately return False on the first unsuccessful
                       action and not execute any remaining actions,
                       defaults to True.
    :type early_stop: bool, optional
    """

    def __init__(self,
                 actions: List[BoboAction],
                 all_success: bool = True,
                 early_stop: bool = True) -> None:
        super().__init__()

        self._actions = actions
        self._all_success = all_success
        self._early_stop = early_stop

    def perform_action(self, event: CompositeEvent) -> bool:
        results = []

        for action in self._actions:
            if action.perform_action(event):
                results.append(True)
            elif self._early_stop:
                return False
            else:
                results.append(False)

        return all(r is True for r in results) if self._all_success \
            else any(r is True for r in results)



import json

import pika

import bobocep.setup.distributed.bobo_dist_constants as bdc
from bobocep.decider.dist_decider import DistDecider
from bobocep.rules.bobo_rule_builder import BoboRuleBuilder
from bobocep.setup.distributed.incoming.dist_incoming_subscriber import \
    IDistIncomingSubscriber
from bobocep.setup.distributed.outgoing.dist_outgoing_subscriber import \
    IDistOutgoingSubscriber
from bobocep.setup.task.bobo_task import BoboTask


class BoboDistIncoming(BoboTask,
                       IDistOutgoingSubscriber):
    """Handles incoming synchronisation data from other :code:`bobocep`
    instances.

    :param decider: The decider to which synchronisation will occur.
    :type decider: DistDecider

    :param exchange_name: The exchange name to connect to on the external
                          message queue system.
    :type exchange_name: str

    :param user_id: The user ID to use on the external message queue system.
    :type user_id: str

    :param host_name: The host name of the external message queue system.
    :type host_name: str
    """

    def __init__(self,
                 decider: DistDecider,
                 exchange_name: str,
                 user_id: str,
                 host_name: str) -> None:

        super().__init__()

        connection = pika.BlockingConnection(
            pika.ConnectionParameters(host=host_name))
        channel = connection.channel()

        channel.exchange_declare(exchange=exchange_name,
                                 exchange_type='direct')

        result = channel.queue_declare('', exclusive=True)
        queue_name = result.method.queue

        channel.queue_bind(exchange=exchange_name, queue=queue_name,
                           routing_key=bdc.TRANSITION)
        channel.queue_bind(exchange=exchange_name, queue=queue_name,
                           routing_key=bdc.CLONE)
        channel.queue_bind(exchange=exchange_name, queue=queue_name,
                           routing_key=bdc.HALT)
        channel.queue_bind(exchange=exchange_name, queue=queue_name,
                           routing_key=bdc.FINAL)
        channel.queue_bind(exchange=exchange_name, queue=queue_name,
                           routing_key=bdc.SYNC_REQ)
        channel.queue_bind(exchange=exchange_name, queue=queue_name,
                           routing_key=bdc.SYNC_RES)

        channel.basic_consume(queue=queue_name,
                              on_message_callback=self._callback,
                              auto_ack=True)

        self.decider = decider
        self.exchange_name = exchange_name
        self.user_id = user_id
        self.host_name = host_name

        self._subs = []
        self._connection = connection
        self._channel = channel
        self._is_synced = False

    def _loop(self) -> None:
        self._channel.start_consuming()

    def _cancel(self) -> None:
        self._is_synced = False
        self._channel.stop_consuming()
        self._connection.close()

    def on_sync(self) -> None:
        if not self._cancelled:
            self._is_synced = True

    def subscribe(self, subscriber: IDistIncomingSubscriber) -> None:
        """
        :param subscriber: Subscribes to incoming data.
        :type subscriber: IDistIncomingSubscriber
        """
        if not self._cancelled:
            if subscriber not in self._subs:
                self._subs.append(subscriber)

    def unsubscribe(self, unsubscriber: IDistIncomingSubscriber) -> None:
        """
        :param unsubscriber: Unsubscribes to incoming data.
        :type unsubscriber: IDistIncomingSubscriber
        """

        if unsubscriber in self._subs:
            self._subs.remove(unsubscriber)

    def _callback(self, ch, method, properties, body):
        if properties.message_id == self.user_id:
            return

        if method.routing_key == bdc.TRANSITION:
            self._handle_transition(body)

        elif method.routing_key == bdc.CLONE:
            self._handle_clone(body)

        elif method.routing_key == bdc.HALT:
            self._handle_halt(body)

        elif method.routing_key == bdc.FINAL:
            self._handle_final(body)

        elif method.routing_key == bdc.SYNC_REQ:
            self._handle_sync_request(ch, method, properties, body)

        elif not self._is_synced and method.routing_key == bdc.SYNC_RES:
            self._handle_sync_response(ch, method, properties, body)

    def _handle_transition(self, data: str) -> None:
        json_data = json.loads(data)
        nfa_name = json_data[bdc.NFA_NAME]
        run_id = json_data[bdc.RUN_ID]
        state_from = json_data[bdc.STATE_FROM]
        state_to = json_data[bdc.STATE_TO]
        event = BoboRuleBuilder.event(json_data[bdc.EVENT])

        for subscriber in self._subs:
            subscriber.on_dist_run_transition(
                nfa_name=nfa_name,
                run_id=run_id,
                state_name_from=state_from,
                state_name_to=state_to,
                event=event
            )

    def _handle_clone(self, data: str) -> None:
        json_data = json.loads(data)
        nfa_name = json_data[bdc.NFA_NAME]
        run_id = json_data[bdc.RUN_ID]
        state_to = json_data[bdc.STATE_TO]
        event = BoboRuleBuilder.event(json_data[bdc.EVENT])

        for subscriber in self._subs:
            subscriber.on_dist_run_clone(
                nfa_name=nfa_name,
                run_id=run_id,
                next_state_name=state_to,
                next_event=event
            )

    def _handle_halt(self, data: str) -> None:
        json_data = json.loads(data)
        nfa_name = json_data[bdc.NFA_NAME]
        run_id = json_data[bdc.RUN_ID]

        for subscriber in self._subs:
            subscriber.on_dist_run_halt(
                nfa_name=nfa_name,
                run_id=run_id
            )

    def _handle_final(self, data: str) -> None:
        json_data = json.loads(data)
        nfa_name = json_data[bdc.NFA_NAME]
        run_id = json_data[bdc.RUN_ID]
        history = BoboRuleBuilder.history(json_data[bdc.HISTORY])

        for subscriber in self._subs:
            subscriber.on_dist_run_final(
                nfa_name=nfa_name,
                run_id=run_id,
                history=history
            )

    def _handle_sync_request(self, ch, method, properties, body):
        if properties.message_id == self.user_id:
            return

        json_data = json.dumps(self.decider.get_current_state())

        ch.basic_publish(exchange=self.exchange_name,
                         routing_key=properties.reply_to,
                         properties=pika.BasicProperties(
                             message_id=self.user_id,
                             correlation_id=properties.correlation_id),
                         body=json_data)

    def _handle_sync_response(self, ch, method, properties, body):
        if properties.message_id == self.user_id:
            return

        for subscriber in self._subs:
            subscriber.on_sync_response(properties.correlation_id, body)

    def _setup(self) -> None:
        """"""



from bobocep.receiver.validators.abstract_validator import \
    AbstractValidator


class AnyValidator(AbstractValidator):
    """A validator that accepts any data."""

    def __init__(self) -> None:
        super().__init__()

    def validate(self, data) -> bool:
        return True



from abc import ABC, abstractmethod

from bobocep.rules.events.primitive_event import PrimitiveEvent


class IReceiverSubscriber(ABC):
    """An interface to subscribe to Receiver events."""

    @abstractmethod
    def on_receiver_event(self, event: PrimitiveEvent) -> None:
        """
        When a Receiver has generated a new PrimitiveEvent instance.

        :param event: A new event.
        :type event: BoboEvent
        """

    @abstractmethod
    def on_invalid_data(self, data) -> None:
        """
        When a Receiver rejects data because it does not pass validity checks.

        :param data: Invalid data.
        :type data: any
        """



from abc import abstractmethod
from threading import RLock


class BoboTask:
    """A task that is scheduled to repeat indefinitely until cancelled."""

    def __init__(self) -> None:
        super().__init__()

        self._setup_done = False
        self._cancelled = False
        self._lock = RLock()

    @abstractmethod
    def _setup(self) -> None:
        """"""

    @abstractmethod
    def _loop(self) -> None:
        """"""

    @abstractmethod
    def _cancel(self) -> None:
        """"""

    def setup(self) -> None:
        """Perform initial setup. It can only be performed once, and must be
        performed before loop functionality can occur.

        :raises RuntimeError: Task has already been cancelled.
        :raises RuntimeError: Setup has already been performed.
        :raises RuntimeError: Exception raised during setup.
        """

        with self._lock:
            if self._cancelled:
                raise RuntimeError("Task has already been cancelled.")

            if self._setup_done:
                raise RuntimeError("Setup has already been performed.")

            try:
                self._setup()
            except Exception as e:
                raise RuntimeError("Exception raised during setup: {}."
                                   .format(e))
            self._setup_done = True

    def loop(self) -> None:
        """Perform primary task loop. The task must not be cancelled, and
        setup must have already occurred.

        :raises RuntimeError: Task has already been cancelled.
        :raises RuntimeError: Setup must be performed before using loop.
        :raises RuntimeError: Exception raised during loop.
        """

        with self._lock:
            if self._cancelled:
                raise RuntimeError("Task has already been cancelled.")

            if not self._setup_done:
                raise RuntimeError(
                    "Setup must be performed before using loop.")

            try:
                self._loop()
            except Exception as e:
                raise RuntimeError("Exception raised during loop: {}."
                                   .format(e))

    def cancel(self) -> None:
        """Cancel the task. This will prevent setup and loop from being
        called.

        :raises RuntimeError: Task has already been cancelled.
        :raises RuntimeError: Exception raised during cancel.
        """

        with self._lock:
            if self._cancelled:
                raise RuntimeError("Task has already been cancelled.")

            try:
                self._cancel()
            except Exception as e:
                raise RuntimeError("Exception raised during cancel: {}."
                                   .format(e))

            self._cancelled = True

    def is_cancelled(self) -> bool:
        """
        :return: True if the task has been cancelled, False otherwise.
        """

        with self._lock:
            return self._cancelled
